open Spotlib.Spot

(** {6 Display machine with Format like capability } *)
       
module Make(A : sig
  type t with conv(ocaml_of)
  val format : (Format.t -> unit) (** printer for the internal *)
               -> Format.t -> t -> unit 
  val buffer : (Buffer.t -> unit) (** buffer output for the internal *)
               -> Buffer.t -> t -> unit
end) : sig

  module Token : sig
    type t =
      | String of string
      | Box of int * t  (** "@[<n>...@]" *)
      | VBox of int * t (** "@[<vn>...@]" *)
      | Cut             (** "@," (known also as Good Break) *)
      | Space           (** "@ " *)
      | Flush           (** "@." *)
      | Seq of t list
      | NOP
      | Attr of A.t * t (** ignored by [format], [dump], [buffer] and [show] *) 

    (** {6 Printer to strings } 

     They simply ignore attributes.
     *)
                        
    val format : Format.t -> t -> unit
    val dump   : Format.t -> t -> unit
    val buffer : Buffer.t -> t -> unit
    val show   : t -> string
  end
  
  type token = Token.t

  (** Primitives *)
  type assoc = Left | Right | Noassoc
  type level = float
  
  type 'a m = assoc -> level -> 'a
  module Monad : Monad_intf.T with type 'a t = 'a m
  
  type t = token m
  
  val box    : int -> t -> t
  val vbox   : int -> t -> t
  val ( ++ ) : t -> t -> t
  val cut    : t
  val space  : t
  val flush  : t
  val seq    : t list -> t
  val string : string -> t
  val nop    : t
  val attr : A.t -> t -> t
  
  (** {6 Level and environments} *)
  
  val left    : 'a m -> 'a m           
  (** Put the arg in the "left" environment *)
  val right   : 'a m -> 'a m          
  (** Put the arg in the "right" environment *)
  val noassoc : 'a m -> 'a m        
  (** Put the arg in the "noassoc" environment *)
  val level   : level -> 'a m -> 'a m 
  (** Put the arg in the specified level *)
  val reset   : 'a m -> 'a m          
  (** Put the arg in "noassoc" env and 0.0 level *)
  (* CR jfuruse: Should be -1.0 ? *)
  
  val check_against_current_level : level -> [ `Same | `Stronger | `Weaker ] m
  (** Compare the argument against the current level:
      `Same     : the same as the current level
      `Stronger : the argument is stronger than the current
      `Weak     : the argument is weaker than the current
  *)
  
  val need_paren : assoc -> level -> bool m
  (** Check the printer object of the level [level] requires parenthesis wrapping or not *)
  
  (** Utilities 
      These implements common printing with auto parenthesis wrapping.
      Note: They may not suit with your own taste.
  *)

  val parenbox : assoc -> level -> t -> t
  val binop    : assoc -> level -> op:t -> t -> t -> t
  val list     : level -> t -> t list -> t
  (** [list lev sep ts] is to create a list without a box *)
  
  val prefix   : level -> op:t -> t -> t
  val postfix  : level -> op:t -> t -> t
  
  val parens : string -> string -> t -> t
  (** Surround the given t with the left and right strings.
      The level for the internal t is reset to -1.0 to prevent from having auto parens for it *)
  
  (** OCaml like printers *)
  module OCaml : sig
    val sequence     : t list -> t
    val if_then_else : t -> t -> t -> t
    val if_then      : t -> t -> t
    val ty_as         : t -> t -> t (** for [t as 'a] *)
    val tuple        : t list -> t
    val ( ^-> )      : t -> t -> t
    val ( + )        : t -> t -> t
    val ( - )        : t -> t -> t
    val ( * )        : t -> t -> t
    val ty_tuple     : t list -> t
    val uminus       : t -> t
    val app          : t -> t -> t
  end
  
  (** Drivers *)
  val buffer : ('a -> t) -> Buffer.t -> ?assoc:assoc -> ?level:level -> 'a -> unit
  val show : ('a -> t) -> ?assoc:assoc -> ?level:level -> 'a -> string
  val format : ?assoc:assoc -> ?level:level -> ('a -> t) -> Format.t -> 'a -> unit
  (** [format] has a different type than [buffer] and [show], 
      so that it could be used smoothly in [Format.fprintf]'s functional setting.
  *)
  
  module MakeDrivers(M : sig type t val ppr : t -> token m end) : sig
    open M
    val format : Format.t -> t -> unit
    val dump   : Format.t -> t -> unit
    val buffer : Buffer.t -> ?assoc:assoc -> ?level:level -> t -> unit
    val show   : ?assoc:assoc -> ?level:level -> t -> string
  end
  
  module Test :sig 
    val test : unit -> unit
  end
end
